#!/usr/bin/env python
import argparse
import logging as LOG
import yaml
from pysnmp.hlapi import *

LOG_FILE = '/var/log/zabbix/zabbix_cbis.log'
LOG_FORMAT = "%(asctime)s.%(msecs).03d %(name)s[%(process)d] %(threadName)s %(levelname)s - %(message)s"
LOG_DATE_FMT = "%Y-%m-%d %H:%M:%S"


# Configuration files
CONSUMERS_FILE = '/etc/cbis-snmp/snmp_trap_dests.yaml'
OID_MAPPING_FILE = '/etc/cbis-snmp/zabbix_oid_mapping.yaml'

# cbis oids
NOKIA_PRODUCTS = '1.3.6.1.4.1.94.1'
CB_PRODUCTS = NOKIA_PRODUCTS + '.169'
CBIS = CB_PRODUCTS + '.1'
CBIS_ALARMS = CBIS + '.1'
CBIS_ALARM_OBJECTS = CBIS_ALARMS + '.1'
CBIS_ALARM_PREFIX = CBIS_ALARMS + '.2'
CBIS_ZABBIX_ALARM_PREFIX = CBIS_ALARM_PREFIX + '.1'

# cbis alarm object to oid
ALARM_OBJECTS = {
    'eventid': CBIS_ALARM_OBJECTS + '.1',
    'rawtext': CBIS_ALARM_OBJECTS + '.2',
    'message': CBIS_ALARM_OBJECTS + '.3',
    'hostId': CBIS_ALARM_OBJECTS + '.4',
    'hostType': CBIS_ALARM_OBJECTS + '.5',
    'hostName': CBIS_ALARM_OBJECTS + '.6',
    'eventTime': CBIS_ALARM_OBJECTS + '.7',
    'thresholdInfo': CBIS_ALARM_OBJECTS + '.8'
}

CBIS_ALARM_STATE_OK = '.1'
CBIS_ALARM_STATE_WARNING = '.2'
CBIS_ALARM_STATE_CRITICAL = '.3'

SEVERITY_MAP = {
    '1': CBIS_ALARM_STATE_WARNING,
    '2': CBIS_ALARM_STATE_WARNING,
    '3': CBIS_ALARM_STATE_WARNING,
    '4': CBIS_ALARM_STATE_CRITICAL,
    '5': CBIS_ALARM_STATE_CRITICAL
}


def load_yaml_file(full_path, with_exception=False):
    with open(full_path, 'r') as stream:
        try:
            return yaml.load(stream, Loader=yaml.BaseLoader)
        except Exception as e:
            LOG.error("[CBIS] Failed to load file: %s. %s" % (full_path, e))
            if with_exception:
                raise e
            else:
                return None


def extract_zabbix_alert_values(body):

    alert_args = {}
    for line in body.split('\n'):
        if '=' not in line:
            continue
        try:
            macro, value = line.split('=', 1)
        except ValueError, e:
            LOG.debug('[CBIS] skipped line %s. Problem: %s', line, e)
            continue

        alert_args[macro.strip()] = value.strip()
    return alert_args


def get_var_binds(alert_values):

    var_binds = [(ALARM_OBJECTS[key], OctetString(alert_values.get(key, 'NA')))
                 for key in ALARM_OBJECTS.keys()]

    return var_binds


def send_snmp_trap(host, var_binds, alert_oid, severity_oid):

    host_details = host['host']

    send_to = str(host_details['send_to'])
    port = str(host_details.get('port', 162))
    community_str = host_details.get('community', 'public')

    LOG.debug("[CBIS] trap parameters: send_to: %s, port: %s, community string: %s" % (send_to, port, community_str))

    error_indication, error_status, error_index, var_bins = next(
        sendNotification(
            SnmpEngine(),
            CommunityData(community_str, mpModel=1),
            UdpTransportTarget((send_to, port)),
            ContextData(),
            'trap',
            NotificationType(
                ObjectIdentity(CBIS_ZABBIX_ALARM_PREFIX + alert_oid + severity_oid),
                ).addVarBinds(*var_binds)
        )
    )

    if error_indication:
        LOG.debug('[CBIS] Notification not sent: %s' % error_indication)
    elif error_status:
        LOG.debug('[CBIS] Notification Receiver returned error: %s @%s' %
                  (error_status, error_index))


def get_severity_oid(alert_values):

    alarm_state = str(alert_values.get('alarmState', None))

    if alarm_state == 'OK':
        return CBIS_ALARM_STATE_OK
    elif alarm_state == 'PROBLEM':
        severity_value = str(alert_values.get('severityValue', '-1'))

        if severity_value in SEVERITY_MAP:
            return SEVERITY_MAP[severity_value]
        else:
            severityText = str(alert_values.get('severityValue', None))
            LOG.debug('[CBIS] Unsupported severity. Severity: %s', severityText)
            return None
    else:
        LOG.debug('[CBIS] Unsupported alarm state: %s' % alarm_state)
        return None


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument("summary", help="alert summary")
    parser.add_argument("body", help="alert body")
    args = parser.parse_args()

    LOG.debug("[CBIS] summary=%s, body=%s" % (args.summary, args.body))

    hosts = load_yaml_file(CONSUMERS_FILE, True)
    alerts_details = load_yaml_file(OID_MAPPING_FILE, True)

    alert_values = extract_zabbix_alert_values(args.body)
    rawtext = alert_values.get('rawtext', None)

    alert_severity_oid = get_severity_oid(alert_values)

    if not alert_severity_oid:
        return

    alert_details = None
    if rawtext:
        alert_details = alerts_details.get(rawtext, None)


    if alert_details:
        for host in hosts:
            LOG.debug("[CBIS] host: %s" % str(host))
            send_snmp_trap(host,
                           get_var_binds(alert_values),
                           alert_details['oid'],
                           alert_severity_oid)
    else:
        LOG.debug('[CBIS] Unregconized alarm. Alarm type: %s', alert_values['rawtext'])


if __name__ == '__main__':

    LOG.basicConfig(filename=LOG_FILE, format=LOG_FORMAT, datefmt=LOG_DATE_FMT, level=LOG.DEBUG)
    main()
