import argparse
import unicodecsv as csv
import datetime
import os
import re
import sys
import time
from pyzabbix import ZabbixAPI
from collections import defaultdict
import socket


INPROGRESS = '_inprogress'
NAME = 'name'
KEY = 'key_'
DISCOVERED = u'4'
NET_RULE_KEY = 'net.if.discovery'
VFS_RULE_KEY = 'vfs.fs.discovery'
NETWORK_APPLICATION = 'Network interfaces'
FILESYSTEM_APPLICATION = 'Filesystems'
HARDWARE_APPLICATION = 'Hardware'
NET_CONNECTORS = 'Net Connectors'
SERVER_BLADES = 'Server Blades'
NETWORKING = 'Networking'
POWER_SUPPLIES = 'Power Supplies'
COMMON_ENCLOSURE_MANAGER='Common Enclosure Manager'
APPLICATIONS_FOR_SUBSTR = [NETWORK_APPLICATION, FILESYSTEM_APPLICATION, HARDWARE_APPLICATION,
                           NET_CONNECTORS, SERVER_BLADES, NETWORKING, POWER_SUPPLIES, COMMON_ENCLOSURE_MANAGER]
GENERAL_COMPUTE = 'general_compute'
OTHERS_CONTROLLER = 'others_controller'
ICMP_COMPUTE = 'icmp_compute'
OS_COMPUTE = 'os_compute'
PROCESSES_COMPUTE = 'processes_compute'
CPU_COMPUTE = 'cpu_compute'
MEMORY_COMPUTE = 'memory_compute'
NETWORK_INTERFACES_COMPUTE = 'network interfaces_compute'
CEPH_CONTROLLER = 'ceph_controller'
GENERAL_STORAGE = 'general_storage'
GENERAL_CONTROLLER = 'general_controller'
ICMP_STORAGE = 'icmp_storage'
ICMP_CONTROLLER = 'icmp_controller'
OS_STORAGE = 'os_storage'
OS_CONTROLLER = 'os_controller'
PROCESSES_STORAGE = 'processes_storage'
PROCESSES_CONTROLLER = 'processes_controller'
CPU_STORAGE = 'cpu_storage'
CPU_CONTROLLER = 'cpu_controller'
MENORY_STORAGE = 'memory_storage'
MEMORY_CONTROLLER = 'memory_controller'
NETWORK_INTERFACES_STORAGE = 'network interfaces_storage'
NETWORK_INTERFACES_CONTROLLER = 'network interfaces_controller'
RETAINED_METRICS = [GENERAL_COMPUTE, OTHERS_CONTROLLER, ICMP_COMPUTE, OS_COMPUTE, PROCESSES_COMPUTE,
                    CPU_COMPUTE, MEMORY_COMPUTE, NETWORK_INTERFACES_COMPUTE, CEPH_CONTROLLER,
                    GENERAL_STORAGE, GENERAL_CONTROLLER, ICMP_STORAGE, ICMP_CONTROLLER, OS_STORAGE,
                    OS_CONTROLLER, PROCESSES_STORAGE, PROCESSES_CONTROLLER, CPU_STORAGE,
                    CPU_CONTROLLER, MENORY_STORAGE, MEMORY_CONTROLLER, NETWORK_INTERFACES_STORAGE,
                    NETWORK_INTERFACES_CONTROLLER]

def get_lock(process_name):
    global lock_socket
    lock_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    try:
        lock_socket.bind('\0' + process_name)
    except socket.error:
        print('Failed to start, another instance is already running.')
        sys.exit()
					
class Host_app_metrics(object):
    def __init__(self, export_function, timestamp = '', measurement_type = '', host_name = '', host_id = '', host_ip = ''):
        self.export_func = export_function
        self.row = []
        self.measurement_type = measurement_type
        self.host_name = host_name
        self.host_id = host_id
        self.host_ip = host_ip
        self.timestamp = timestamp
        self.metrics_list = {}

    def add_metric(self, metric_name = '', metric_value = ''):
        self.metrics_list[metric_name] = metric_value

    def print_header(self, keys=None):
        row_to_print = []
        self._append_to_row(row=row_to_print, data='measurement_type')
        self._append_to_row(row=row_to_print, data='timestamp')
        self._append_to_row(row=row_to_print, data='host_id')
        self._append_to_row(row=row_to_print, data='host_name')
        self._append_to_row(row=row_to_print, data='host_ip')

        if keys:
            for key in sorted(keys):
                self._append_to_row(row=row_to_print, data=key)
        else:
            for key in (sorted(self.metrics_list.keys())):
                self._append_to_row(row=row_to_print, data=key)
        self.export_func(row_to_print)

    def print_line(self, keys=None):
        row_to_print = []
        self._append_to_row(row=row_to_print, data=self.measurement_type + '_' + self.extract_host_type(host_name=self.host_name))
        self._append_to_row(row=row_to_print, data=self.timestamp)
        self._append_to_row(row=row_to_print, data=self.host_id)
        self._append_to_row(row=row_to_print, data=self.host_name)
        self._append_to_row(row=row_to_print, data=self.host_ip)

        if keys:
            for key in (sorted(keys)):
                self._append_to_row(row=row_to_print,
                                    data=self.metrics_list.get(key) if self.metrics_list.get(key) else '')
        else:
            for key in (sorted(self.metrics_list.keys())):
                self._append_to_row(row=row_to_print, data=self.metrics_list.get(key))

        self.export_func(row_to_print)

    @staticmethod
    def measurement_type_host_name(measurement_type=None, host_name=None):
        return measurement_type + '_' + Host_app_metrics.extract_host_type(host_name=host_name)

    @staticmethod
    def extract_host_type(host_name):
        if not host_name:
            return 'unknown'
        type_list = ['compute', 'controller', 'storage', 'undercloud', 'airframe', 'enclosure']
        for host_type in type_list:
            if host_type in host_name.lower():
                return host_type
        return 'unknown'

    def _append_to_row(self, row, data):
        if data:
            row.append(data)
        else:
            row.append('')

def get_zabbix_client(url, user, passwd, verify):
    """
    :param url: Zabbix API url
    :param user: Zabbix API user
    :param passwd: Zabbix API password
    :param verify: Whether or not to verify the SSL certificate
    :return: ZABBIX API Client
    """
    client = ZabbixAPI(url)
    client.session.verify = verify
    client.login(user, passwd)

    return client


def get_csv_writer(output_file):
    """

    :param output_file: csv output file path
    :return: csv writer to output file
    """
    return csv.writer(open(output_file, "wb+")).writerow


def remove_old_version(directory, date_to_remove_from, time_in_day):
    """

    :param directory: directory to remove from.
    :param date_to_remove_from: date to remove file from.
    :param time_in_day: remove corresponding metrics file to the time in day in the above date.
    """
    remove_date = datetime.date.fromordinal(datetime.date.today().toordinal() - date_to_remove_from)
    remove_file = os.path.join(directory, 'zbx_metrics_' + remove_date.strftime('%d%m%Y') + '_' + time_in_day + '.csv')
    if os.path.exists(remove_file):
        os.remove(remove_file)


def build_parser():
    """
    Builds the argparser object
    :return: Configured argparse.ArgumentParser object
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Zabbix Metrics Exporter')
    parser.add_argument('-V', '--verify',
                        default='True',
                        choices=['True', 'False'],
                        help='Verify SSL (True, False)')
    parser.add_argument('-U', '--url',
                        default='https://10.161.6.7/zabbix/',
                        help='Zabbix API server'
                             'example: https://zabbixhost.example.com/zabbix')
    parser.add_argument('-u', '--user',
                        default='metrics-exporter',
                        help='Zabbix API user')
    parser.add_argument('-p', '--passwd',
                        default='zabbix',
                        help='Zabbix API password')
    parser.add_argument('-m', '--minutes-ago',
                        default='65',
                        type=int,
                        help='How many minutes worth of metrics should'
                             'be returned going back in time from right now')
    parser.add_argument('-o', '--odir',
                        default='/var/log/zabbix/metrics/',
                        help='Output directory to export the csv to')
    parser.add_argument('-i', '--itemid',
                        help='Zabbix item id to metrics export')
    parser.add_argument('-H', '--hostid',
                        help='Zabbix host id to metrics export')
    parser.add_argument('-t', '--type',
                        choices=(0, 3),
                        default=3,
                        help='Zabbix item value type:\n'
                             '0: numeric float\n'
                             '3: numeric unsigned\n')
    parser.add_argument('-k', '--keep',
                        default=1,
                        help='Days to keep metrics files for.')

    return parser


def _print_func(x):
    print(x)


def export_metrics(z_client, period=65, hostsids=None, itemids=None, value_type=3, export_func=_print_func):
    """

    :param z_client: Zabbix Client
    :param period: Time period from now to get metrics for(in Minutes)
    :param hostsids: Host ids to query for
    :param itemids: Item ids to query for
    :param value_type: Zabbix item value type 0:numeric float, 3: unsigned float
    :param export_func: The export function used
    """
    # export_func(('key', 'timestamp', 'hostID', 'hostname', 'templateItemID', 'itemID',
    #             'name', 'units', 'value', 'error', 'application'))
    hosts = z_client.host.get(hostids=hostsids, output=['name'])
    macros = z_client.usermacro.get(output='extend', globalmacro=True)
    app_keys = defaultdict(list)
    app_hosts = defaultdict(list)
    all_host_metrics = defaultdict(Host_app_metrics)

    export_func(["#Report period start: " + time.strftime('%Y-%m-%d %H:%M:%S%z')])

    for host in hosts:
        h_id = host['hostid']
        items = z_client.item.get(hostids=h_id,
                                  itemids=itemids,
                                  monitored=True,
                                  selectDiscoveryRule='extend',
                                  selectItemDiscovery='extend',
                                  selectApplications='extend',
                                  output=[NAME, KEY, 'units', 'templateid',
                                          'flags', 'error', 'value_type', 'applications'])
        for item in items:
            i_id = item['itemid']
            i_templateid = item['templateid']
            i_name = item[NAME]
            i_key = item[KEY]
            i_application = 'others'
            if item['applications']:
                i_application = item['applications'][0]['name']
            if not i_application in app_keys:
                app_keys[i_application] = []
                app_hosts[i_application] = [] 
               
            # Item key replacement
            for mac in macros:
                macro_name = mac['macro']
                macro_value = mac['value']
                if macro_name in i_key:
                    i_key = i_key.replace(macro_name,macro_value)

            # Item name replacement
            if "[" in i_key and "$" in i_name:
                params = i_key[i_key.find("[")+1:i_key.find("]")]
                params_array = params.split(',')

                for i in xrange(0,len(params_array)):
                    i_name = re.sub(r"\$" + str(i+1) + r"\b",params_array[i],i_name)
			
            i_units = item['units']
            i_is_discovered = item['flags'] == DISCOVERED
            if_name = ''
            i_type = item.get('value_type', value_type)
            if i_is_discovered:
                i_discovery_rule = item['discoveryRule']
                item_discovery_key = item['itemDiscovery'][KEY]
                if i_application in APPLICATIONS_FOR_SUBSTR:
                    params = i_key[i_key.find("[")+1:i_key.find("]")]
                    if_name = params.split(',')[0]
                    after_first_braket = i_key[i_key.find("[")+1:len(i_key)]
                    if after_first_braket.find(",") >= 0 and after_first_braket.find("]") >= after_first_braket.find(",") :
                        after_first_braket = after_first_braket[after_first_braket.find(",") + 1: len(after_first_braket)]
                    else:
                        after_first_braket = after_first_braket[after_first_braket.find("]"):len(after_first_braket)]
                    i_key = i_key[0: i_key.find("[") + 1] + after_first_braket

            if i_key.startswith('snmptrap'):
                continue
            if not i_key in app_keys.get(i_application):
                app_keys.get(i_application).append(i_key)
            h_name = '{0}{1}'.format(host[NAME], ('[{0}]'.format(if_name) if if_name else ''))
            name_app = h_name + '_' + i_application

            if not Host_app_metrics.measurement_type_host_name(measurement_type=i_application, host_name=h_name).lower() in RETAINED_METRICS:
                # print 'ignoring ' + Host_app_metrics.measurement_type_host_name(measurement_type=i_application, host_name=h_name).lower()
                continue

            if not name_app in all_host_metrics:
               all_host_metrics[name_app] = Host_app_metrics(export_function=export_func, timestamp=time.strftime('%d/%m/%Y %H:%M'),
                                                             host_name=h_name, host_id=h_id, host_ip='1.1.1.1',
                                                             measurement_type=i_application)
               app_hosts[i_application].append(name_app)

            if i_type in ['0','3']:
                i_metrics = [float(metric['value']) for metric in
                             z_client.history.get(itemids=i_id,
                                                  history=i_type,
                                                  time_from=time.time() - 60 * period)]
                i_error = item['error']
                all_host_metrics[name_app].add_metric(metric_name=i_key,
                                                      metric_value=sum(i_metrics)/len(i_metrics) if i_metrics else '')
            else:
                i_metrics = [str(metric['value']) for metric in
                             z_client.history.get(itemids=i_id,
                                                  history=i_type,
                                                  time_from=time.time() - 60 * period)]
                i_error = item['error']
                if i_metrics:
                    i_metrics = [i_metrics[0]]
                all_host_metrics[name_app].add_metric(metric_name=i_key,
                                                      metric_value=i_metrics if i_metrics else '')


    export_func((''))
    export_func(["#Report period end: " + time.strftime('%Y-%m-%d %H:%M:%S%z')])
    if app_hosts:
        for application_key in sorted(app_hosts.keys()):
            export_func((''))
            for name_app_key in app_hosts.get(application_key):
                all_host_metrics[name_app_key].print_header(keys=app_keys.get(application_key));
                break;
            for name_app_key in sorted(app_hosts.get(application_key)):
                all_host_metrics[name_app_key].print_line(keys=app_keys.get(application_key));


if __name__ == '__main__':
    arg_parser = build_parser()
    args = arg_parser.parse_args(sys.argv[1:])
    now = time.strftime('%H%M')
    get_lock('zbx_metrics_exporter')

    try:
        z_client = get_zabbix_client(url=args.url, user=args.user, passwd=args.passwd,
                                     verify=args.verify.strip().lower() == 'true')
    except Exception as e:
        if 'Login name or password is incorrect.' in str(e):
            print('Unauthorized: Please check your username and password')
        else:
            print('Error connecting to zabbixapi: {0}'.format(e))
        exit(1)

    try:
        file_name = os.path.join(args.odir, 'zbx_metrics_' + time.strftime('%d%m%Y') + '_' + now + '.csv')
        csv_writer = get_csv_writer(file_name + INPROGRESS)
        export_metrics(z_client=z_client,
                       hostsids=args.hostid,
                       itemids=args.itemid,
                       export_func=csv_writer
                       )
        os.rename(file_name + INPROGRESS, file_name)
    except Exception as e:
        print('An error has occurred while exporting metrics.\nError:\n{0}'.format(e))
        exit(1)

    try:
        remove_old_version(args.odir, args.keep, now)
    except Exception as e:
        print('An error has occurred while deleting old metric file.\nError:\n{0}'.format(e))
        exit(1)


